/*
    Copyright 2005-2008 Intel Corporation.  All Rights Reserved.

    This file is part of Threading Building Blocks.

    Threading Building Blocks is free software; you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

    Threading Building Blocks is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Threading Building Blocks; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    As a special exception, you may use this file as part of a free software
    library without restriction.  Specifically, if other files instantiate
    templates or use macros or inline functions from this file, or you compile
    this file and link it with other files to produce an executable, this
    file does not by itself cause the resulting executable to be covered by
    the GNU General Public License.  This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.
*/

/*
Class hierarchy:
atomic_word
atomic_base
  atomic_impl1
    atomic for bool and void*
    atomic_impl2 (additive operations)
      atomic for pointers
      atomic_impl3 (bitwise operations)
        atomic for (un)signed integers and character types
*/

#ifndef __TBB_atomic_H
#define __TBB_atomic_H

#include <cstddef>
#include "tbb_stddef.h"

#include "tbb_machine.h"

#if defined(_MSC_VER) && defined(_Wp64)
    // Workaround for overzealous compiler warnings in /Wp64 mode
    #pragma warning (push)
    #pragma warning (disable: 4244 4267)
#endif /* _MSC_VER && _Wp64 */

namespace tbb {

//! @cond INTERNAL
namespace internal {

// atomic_base

    // Without forced alignment, an 8-byte field may well be aligned on a 4-byte boundary,
    // and alignment is required because: TODO: analysis of whether such alignment assures
    // atomicity on all supported platforms

    #if __GNUC__ || __SUNPRO_CC
    #define __TBB_M(t,f,a) t f __attribute__ ((aligned(a)));
    #elif defined(__INTEL_COMPILER)||_MSC_VER >= 1300
    #define __TBB_M(t,f,a) __declspec(align(a)) t f;
    #else 
    #error Do not know syntax for forcing alignment.
    #endif

    // TODO: how to catch non-specialized uses other than by absence of my_value?
    template<typename T,size_t S> class atomic_base {}; // Primary template

    // TODO: use private with friend declaration instead of protected?
    #define __TBB_MM(S) template<typename T> class atomic_base<T,S> { protected: __TBB_M(T,my_value,S) };
    __TBB_MM(1)
    __TBB_MM(2)
    __TBB_MM(4)
    __TBB_MM(8)
    #undef __TBB_MM

    #undef __TBB_M

// TODO: if typedef's have to be repeated in derived structs, will the atomics have them? do they need them?

// TODO: An atomic should probably have an extra template parameter
// that governs what the default memory semantics are for operations:
// - relaxed for reference counts
// - rel_acq for "normal" release/acquire/rel_acq defaults (it would be the defaults' default)
// - ordered for special purposes
template<typename T>
struct atomic_impl1 : atomic_base<T,sizeof(T)> {
private:
    typedef typename atomic_word<sizeof(T)>::word word;

public:
    typedef T value_type;

    template<memory_semantics M>
    value_type load() const {
        return value_type(__TBB_Load<sizeof(value_type),word,M>::op( &this->my_value ));
    }

    value_type load() const {
        return load<acquire>();
    }

    operator value_type() const {
        return load();
    }

    template<memory_semantics M>
    void store( value_type value ) {
        __TBB_Op<sizeof(value_type),word,M,__TBB_op_store>::op( &this->my_value, word(value) );
    }

    void store( value_type value ) {
        store<release>(value);
    }

    // operator= defined in atomic classes (TODO: any particular reason?)

    template<memory_semantics M>
    value_type compare_and_swap( value_type value, value_type comparand ) {
        // TODO: check again that this form of casting also does the ?1:0 thing for bool types
        // TODO: sizeof(bool) or sizeof(char)?
        return value_type(__TBB_CompareAndSwap<sizeof(value_type),word,M>::op(&this->my_value,word(value),word(comparand)));
    }

    value_type compare_and_swap( value_type value, value_type comparand ) {
        return compare_and_swap<rel_acq>(value,comparand);
    }

    template<memory_semantics M>
    value_type fetch_and_store( value_type value ) {
        return value_type(
            __TBB_FetchOp<sizeof(value_type),word,M,__TBB_op_store>::op( &this->my_value, word(value) )
            );
    }

    value_type fetch_and_store( value_type value ) {
        return fetch_and_store<rel_acq>(value);
    }
};

//! Additive inverse of 1 for type T.
/** Various compilers issue various warnings if -1 is used with various integer types.
    The baroque expression below avoids all the warnings (we hope). */
#define __TBB_MINUS_ONE(T) (T(T(0)-T(1)))

// This type adds additive operations.
template<typename T, typename D, size_t Step>
struct atomic_impl2: atomic_impl1<T> {
private:
    typedef typename atomic_word<sizeof(T)>::word word;

public:
    typedef T value_type;

    template<memory_semantics M>
    value_type fetch_and_add( D arg ) {
        return value_type(
            __TBB_FetchOp<sizeof(value_type),word,M,__TBB_op_add>
            ::op( &this->my_value, Step*arg )
            );
    }

    value_type fetch_and_add( D arg ) {
        return fetch_and_add<rel_acq>(arg);
    }

    // Additive inverse of arg computed using binary minus,
    // instead of unary minus, for sake of avoiding compiler warnings.
    template<memory_semantics M>
    value_type fetch_and_subtract( D arg ) {
        return fetch_and_add<M>(D(0)-arg);
    }

    value_type fetch_and_subtract( D arg ) {
        return fetch_and_subtract<rel_acq>(arg);
    }

    template<memory_semantics M>
    value_type fetch_and_increment() {
        return fetch_and_add<M>(1);
    }

    value_type fetch_and_increment() {
        return fetch_and_add(1);
    }

    template<memory_semantics M>
    value_type fetch_and_decrement() {
        return fetch_and_add<M>(__TBB_MINUS_ONE(D));
    }

    value_type fetch_and_decrement() {
        return fetch_and_add(__TBB_MINUS_ONE(D));
    }

    value_type& _internal_reference() const {
        return static_cast<value_type&>(this->my_value);
    }

public:
    value_type operator+=( D arg ) {
        return fetch_and_add(arg)+arg;
    }

    value_type operator-=( D arg ) {
        return fetch_and_subtract(arg)-arg;
    }

    value_type operator++() {
        return fetch_and_add(1)+1;
    }

    value_type operator--() {
        return fetch_and_subtract(1)-1;
    }

    value_type operator++(int) {
        return fetch_and_add(1);
    }

    value_type operator--(int) {
        return fetch_and_subtract(1);
    }
};

// This type adds bitwise operations.
template<typename T>
struct atomic_impl3: atomic_impl2<T, T, 1> {
private:
    typedef typename atomic_word<sizeof(T)>::word word;

public:
    typedef T value_type;

    #define __TBB_M(what)                                                    \
        template<memory_semantics M>                                         \
        value_type fetch_and_##what( value_type arg ) {                      \
            return value_type(                                               \
                __TBB_FetchOp<sizeof(value_type),word,M,__TBB_op_##what>     \
                ::op( &this->my_value, arg )                                 \
                );                                                           \
        }                                                                    \
                                                                             \
        value_type fetch_and_##what( value_type arg ) {                      \
            return fetch_and_##what<rel_acq>(arg);                           \
        }
    __TBB_M(and)
    __TBB_M(or )
    __TBB_M(xor)
    #undef __TBB_M

public:
    value_type operator&=( value_type arg ) { return fetch_and_and(arg)&arg; }
    value_type operator|=( value_type arg ) { return fetch_and_or (arg)|arg; }
    value_type operator^=( value_type arg ) { return fetch_and_xor(arg)^arg; }
};

} /* Internal */
//! @endcond

//! Primary template for atomic.
/** See the Reference for details.
    @ingroup synchronization */
template<typename T>
struct atomic {
};

// TODO: why doesn't operator= return a reference to the atomic object?

// atomics based on atomic_impl1

    // TODO question: why not just define operator= once in atomic_impl1?
    #define __TBB_M(T)                                           \
        template<> struct atomic<T>: internal::atomic_impl1<T> { \
            T operator=( T rhs ) { store(rhs); return rhs; }     \
        };

    __TBB_M(bool)
    __TBB_M(void*)

    #undef __TBB_M

    // TODO: no non-scalar types?

// atomics based on atomic_impl2 (includes additive operations)

    template<typename T>
    struct atomic<T*>: internal::atomic_impl2<T*,ptrdiff_t,sizeof(T)> {
        T* operator=( T* rhs ) {
            // "this" required here in strict ISO C++ because store is a dependent name
            this->store(rhs);
            return rhs;
        }
        T* operator->() const {
            return *this;
        }
    };

// atomics based on atomic_impl3 (includes bitwise operations)

    // TODO: why is the difference type not always signed, like ptrdiff_t instead of size_t for atomic<T*>?
    #define __TBB_M(T)                                           \
        template<> struct atomic<T>: internal::atomic_impl3<T> { \
            T operator=( T rhs ) { store(rhs); return rhs; }     \
        };

    // the signed and unsigned integer types
        __TBB_M(  signed char     )
        __TBB_M(unsigned char     )
        __TBB_M(         short    )
        __TBB_M(unsigned short    )
        __TBB_M(         int      )
        __TBB_M(unsigned int      )
        __TBB_M(         long     )
        __TBB_M(unsigned long     )
        #if !( defined(_MSC_VER) && _MSC_VER<1400 && !defined(__INTEL_COMPILER) )
        // "Some old versions of MVSC cannot correctly compile templates with "long long"." (tbb20_20080408oss_src)

        // TODO: OK to have eliminated __TBB_LONG_LONG? The exact same condition was used (also in tbb20_20080408oss_src)
        // to test using "(unsigned) long long", not "__TBB_LONG_LONG", so __int64 (#if _MSC_VER) was never needed?
        // If there are any versions of MVSC that can have an atomic on __int64 but not on long long,
        // this should be reconsidered, of course, with a corresponding test added to test_atomic.cpp.

        __TBB_M(         long long)
        __TBB_M(unsigned long long)
        #endif

    // the other integral types (except bool)
        // TODO: why would these need additive and bitwise operations?
        __TBB_M(char)
        #if !defined(_MSC_VER)||defined(_NATIVE_WCHAR_T_DEFINED) 
        __TBB_M(wchar_t)
        #endif

    // TODO: how about types like the following:
    // C++0x <cstdint> (in namespace std): {,u}int{{,_fast,_least}{8,16,32,64},max,ptr}_t
    // Of these, TBB now already uses {,u}int{8,16,32,64,ptr}_t, but not in std and not for atomics;
    // curiously these are exactly the ones that C++0x marks optional.
    // others proposed for C++0x: {,s}size_t, ptrdiff_t, char16_t, char32_t

    // TODO: any need for shifts and multiplies and maybe others (probably not)?

    #undef __TBB_M

} // namespace tbb

#if defined(_MSC_VER) && defined(_Wp64)
    // Workaround for overzealous compiler warnings in /Wp64 mode
    #pragma warning (pop)
#endif /* _MSC_VER && _Wp64 */

#endif /* __TBB_atomic_H */
