/*
    Copyright 2005-2008 Intel Corporation.  All Rights Reserved.

    This file is part of Threading Building Blocks.

    Threading Building Blocks is free software; you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

    Threading Building Blocks is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Threading Building Blocks; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    As a special exception, you may use this file as part of a free software
    library without restriction.  Specifically, if other files instantiate
    templates or use macros or inline functions from this file, or you compile
    this file and link it with other files to produce an executable, this
    file does not by itself cause the resulting executable to be covered by
    the GNU General Public License.  This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.
*/

#ifndef __TBB_machine_H
#define __TBB_machine_H

#include <cstdio>
#include <assert.h>
#include "tbb/tbb_stddef.h"

// Other file inclusions further down

namespace tbb {

//! Specifies memory semantics/ordering constraint (or what's it also called?).
enum memory_semantics {
    //! Operations on this atomic variable (only) are sequentially consistent.
    // Note that a programmer should probably not feel "relaxed" about using it
    // ("raw" or "unfenced" might be more appropriate to call attention
    // to the lack of implied memory semantics), but "relaxed" has gained a
    // following and it has 7 characters like the other options.
    relaxed,
    //! Additionally, all store operations on any (other) variable by this thread
    //! prior to an operation with release semantics on this atomic variable...
    release, // like StoreStore before the basic operation
    //! ...become visible to loads in any other thread after it performs
    //! an operation with acquire semantics on this (same) variable.
    acquire, // like LoadLoad after the basic operation
    //! Both release and acquire properties.
    // Note that C++0x proposes "acq_rel", but that only makes sense alphabetically,
    // so using rel_acq here is very much intentional.
    rel_acq, // like StoreStore before and LoadLoad after the basic operation
    //! With sequential consistency
    // Note that this is rather heavy, but still the only possibility
    // even if only a StoreLoad fence is needed, e.g., for Dekker's algorithm.
    ordered // like {Store,Load}{Store,Load} before and after the basic operation, right?
};

namespace internal {

    // template metaprogramming
        // And
            template<bool A, bool B> struct And { enum { RET=false }; }; // Primary template
            template<> struct And<true,true> { enum { RET=true }; };
        // Or
            template<bool A, bool B> struct Or { enum { RET=true }; }; // Primary template
            template<> struct Or<false,false> { enum { RET=false }; };
        // If
            template<bool Condition,typename TrueClause,typename FalseClause> struct If {}; // Primary template
            template<typename TrueClause,typename FalseClause> struct If<true ,TrueClause,FalseClause> { typedef TrueClause  RET; };
            template<typename TrueClause,typename FalseClause> struct If<false,TrueClause,FalseClause> { typedef FalseClause RET; };
        // PermanentlyDisabled
            template<bool test> struct PermanentlyDisabled { };
        // Assert
            template<bool test> struct Assert { };
            template<> struct Assert<true> { typedef bool RET; };

    // "allowed" and "test_is_ordered" typedefs meant for use with SFINAE (Substitution Failure Is Not An Error):
    // - "allowed" is at the conceptual level:
    //     - with memory semantics "release"/"acquire"/"rel_acq", the user shall not try combinations that are not allowed
    //         (TBB will not implement any and will therefore cause the compiler to reject such attempts),
    //         and if allowed TBB will add whatever LoadLoad or StoreStore fences are required
    //         (for many architectures this will only require a compiler-level fence, for Itanium with Intel compiler not even that?);
    //         a special situation is __TBB_CompareAndSwap, which will assume success and thus pretend to include a store
    //         (on failure, __TBB_fence_for_release will be issued without reason, which does not harm integrity,
    //         but which may or may not imply a performance penalty on non-TSO platforms)
    //         TODO: duplicate semantics, one for success and the other for failure, to get rid of this penalty?
    //     - "ordered" is always allowed
    // - "test_is_ordered" is at the implementation level: TBB will revert to an implementation based
    //     on an ordered basic operation (if one is not present, that would be an porting error),
    //     and the implementation will only add compiler-level total fences before and after the basic operation
    // "non_acquire_to_relaxed" is a hack (unless all template metaprogramming is a hack...), only use macro at end
    template<memory_semantics M> struct memory_semantics_traits {}; // Primary template
    template<> struct memory_semantics_traits<relaxed> {
        template<bool does_store,bool does_load> struct allowed { typedef bool RET; };
        template<bool arg> struct test_is_ordered { typedef bool RET; };
        enum { non_acquire_to_relaxed = relaxed };
    };
    template<> struct memory_semantics_traits<release> {
        template<bool does_store,bool does_load> struct allowed { typedef typename Assert<does_store>::RET RET; };
        template<bool arg> struct test_is_ordered { typedef bool RET; };
        enum { non_acquire_to_relaxed = relaxed };
    };
    template<> struct memory_semantics_traits<acquire> {
        template<bool does_store,bool does_load> struct allowed { typedef typename Assert<does_load>::RET RET; };
        template<bool arg> struct test_is_ordered { typedef bool RET; };
        enum { non_acquire_to_relaxed = acquire };
    };
    template<> struct memory_semantics_traits<rel_acq> {
        template<bool does_store,bool does_load> struct allowed { typedef typename Assert<And<does_store,does_load>::RET>::RET RET; };
        template<bool arg> struct test_is_ordered { typedef bool RET; };
        enum { non_acquire_to_relaxed = relaxed };
    };
    template<> struct memory_semantics_traits<ordered> {
        template<bool does_store,bool does_load> struct allowed { typedef bool RET; };
        template<bool arg> struct test_is_ordered { typedef typename Assert<arg>::RET RET; };
        enum { non_acquire_to_relaxed = relaxed };
    };

    #define __TBB_NON_ACQUIRE_TO_RELAXED(M) \
        ((tbb::memory_semantics)tbb::internal::memory_semantics_traits<M>::non_acquire_to_relaxed)

    // does_store: may be used with release or rel_acq (relates to operation, not to implementation)
    // is_ordered: basic implementation has its own sequential consistency
    // does_load: may be used with acquire or rel_acq (relates to operation, not to implementation)
    // M: desired memory semantics
    template<bool does_store, bool is_ordered, bool does_load, memory_semantics M>
    struct __TBB_fence_guard { // Primary template
        typedef typename PermanentlyDisabled<does_store>::RET RET; // make it "abstract"
    };

} } // namespace tbb::internal

namespace tbb { namespace internal {

// __TBB_CompareAndSwap

    template<size_t S, typename T, memory_semantics M> struct __TBB_CompareAndSwap { // Primary template
        static inline T op (volatile void *ptr, T value, T comparand); // no default definition
    };

    // sizes 1 and 2 have a default definition in terms of size 4
    // that has lower priority than a user-supplied specialization,
    // c/o the separately specified type T
    // (that really should always be atomic_word<S>::word)
    // TODO: consider the idea of finding an alternative means to specify this intermediate
    // priority and then eliminating the 1st or the 2nd template parameter?

// __TBB_(Fetch)Op

    // __TBB_op_op, for generic fetch_op support

        // TODO: would an __TBB_op_subtract be useful, or is it better to just __TBB_op_add a negated value?
        struct __TBB_op_store { template<typename T> static T op(T arg1, T arg2) { return      arg2; } };
        struct __TBB_op_add   { template<typename T> static T op(T arg1, T arg2) { return arg1+arg2; } };
        struct __TBB_op_and   { template<typename T> static T op(T arg1, T arg2) { return arg1&arg2; } };
        struct __TBB_op_or    { template<typename T> static T op(T arg1, T arg2) { return arg1|arg2; } };
        struct __TBB_op_xor   { template<typename T> static T op(T arg1, T arg2) { return arg1^arg2; } };

    template<size_t S, typename T, memory_semantics M, typename atomic_op> struct __TBB_FetchOp { // Primary template
        static inline T op (volatile void *ptr, T arg); // defined further down (using AtomicBackoff)
    };

    // Note: an implementation may mostly benefit from optimising __TBB_op_store,
    // but also __TBB_op_{and,or} as used in __TBB_Atomic{AND,OR}.
    template<size_t S, typename T, memory_semantics M, typename atomic_op> struct __TBB_Op { // Primary template
        // Disallow acquire and rel_acq that __TBB_FetchOp would allow...
        typename memory_semantics_traits<M>::template allowed<true,false>::RET anonymous;
        static inline void op( volatile void *ptr, T arg ) {
            // ...but use its fence anyway (a port can specialise __TBB_Op to avoid that fence and load).
            (void)__TBB_FetchOp<S,T,M,atomic_op>::op( ptr, arg );
        }
    };

// __TBB_Load

    // This will typically be specialized for a particular platform for memory_semantics other than "ordered".
    template<size_t S, typename T, memory_semantics M> struct __TBB_Load { // Primary template
        // Disallow release and rel_acq that __TBB_CompareAndSwap would allow...
        typename memory_semantics_traits<M>::template allowed<false,true>::RET anonymous;
        static inline T op( const volatile void *ptr ) {
            // ...but use its fence anyway (a port can specialise __TBB_Load to avoid that fence).
            // nothing will change if a value is swapped for an identical value
            return __TBB_CompareAndSwap<S,T,M>::op(const_cast<volatile void*>(ptr),42,42);
        }
    };

} } // namespace tbb::internal

// Note that tbb/machine/{linux,mac}_*.h could be called tbb/machine/gcc_*.h,
// because they apply to inline assembler syntax more than to an O.S.

#if _WIN32||_WIN64

    #ifdef _MANAGED
    #pragma managed(push, off)
    #endif

    #if defined(_M_IX86)
    #include "tbb/machine/windows_ia32.h"
    #elif defined(_M_AMD64) 
    #include "tbb/machine/windows_em64t.h"
    #else
    #error Unsupported platform
    #endif

    #ifdef _MANAGED
    #pragma managed(pop)
    #endif

#elif __linux__ || __FreeBSD__

    #if __i386__
    #include "tbb/machine/linux_ia32.h"
    #elif __x86_64__
    #include "tbb/machine/linux_em64t.h"
    #elif __ia64__
    #include "tbb/machine/linux_itanium.h"
    #else
    #error Unsupported platform
    #endif

#elif __APPLE__

    #if __i386__
    #include "tbb/machine/linux_ia32.h"
    #elif __x86_64__
    #include "tbb/machine/linux_em64t.h"
    #elif __POWERPC__
    #include "tbb/machine/mac_ppc.h"
    #else
    #error Unsupported platform
    #endif

#elif _AIX

    #include "tbb/machine/ibm_aix51.h"

#elif __sun || __SUNPRO_CC

    #define __asm__ asm 
    #define __volatile__ volatile 	
    #if __i386  || __i386__
    #include "tbb/machine/linux_ia32.h"
    #elif __x86_64__
    #include "tbb/machine/linux_em64t.h"
    #else
    #error Unsupported platform
    #endif

#else

    #error Unsupported O.S.

#endif

// __TBB_fence_guard

    namespace tbb { namespace internal {

    // TODO: use or eliminate? it does offer some insight to require these even if only as mere documentation, though...
    #if !defined(__TBB_PLATFORM_ORDERS_STORES)
    #error Incomplete information
    #endif
    #if !defined(__TBB_PLATFORM_ORDERS_LOADS)
    #error Incomplete information
    #endif

    // Note that the following invariant check was taken out because it does not need to apply to __TBB_Op<__TBB_op_and>:
    //     "typename Assert<Or<!is_ordered,And<does_store,does_load>::RET>::RET>::RET anonymous1;"
    #define __TBB_M(M,constructor,destructor)                                                                                    \
        template<bool does_store, bool is_ordered, bool does_load> struct __TBB_fence_guard<does_store,is_ordered,does_load,M> { \
            /* some tests */                                                                                                     \
                typename memory_semantics_traits<M>::template allowed<does_store,does_load>::RET anonymous2;                     \
                typename memory_semantics_traits<M>::template test_is_ordered<is_ordered>::RET   anonymous3;                     \
            __TBB_fence_guard() { constructor }                                                                                  \
            ~__TBB_fence_guard() { destructor }                                                                                  \
        };
    __TBB_M(relaxed,                           ,                           )
    __TBB_M(release,__TBB_fence_for_release ();,                           ) // TODO: what if is_ordered?
    __TBB_M(acquire,                           ,__TBB_fence_for_acquire ();) // TODO: what if is_ordered?
    __TBB_M(rel_acq,__TBB_fence_for_release ();,__TBB_fence_for_acquire ();) // TODO: what if is_ordered?
    __TBB_M(ordered,__TBB_fence_for_compiler();,__TBB_fence_for_compiler();)
    #undef __TBB_M

    } } // tbb::internal

#if 0 // TODO: won't work with templates, so devise something else
#if !defined(__TBB_CompareAndSwap4) || !defined(__TBB_CompareAndSwap8) || !defined(__TBB_Yield)
#error Minimal requirements for tbb_machine.h not satisfied 
#endif
#endif

#if defined(TBB_DO_ASSERT)
#define WORDSIZE __TBB_WORDSIZE
#endif

// atomic_word

    namespace tbb { namespace internal {

    template<size_t Size>
    struct atomic_word {             // Primary template
    };

    // Maybe 64-bit Windows needs it, but it's easiest to do this regardless
    // (although the question arises whether int_fastX_t or int_leastX_t is meant
    // for specific uses, so maybe this should be revised again), and meanwhile
    // __TBB_MaskedCompareAndSwap has come to rely on this so any change (back)
    // should be considered carefully.
    /// #if _WIN32 && __TBB_x86_64
    /// // ATTENTION: On 64-bit Windows, we currently have to specialize atomic_word
    /// // for every size to avoid type conversion warnings
    /// // See declarations of atomic primitives in machine/windows_em64t.h
    template<>
    struct atomic_word<1> {          // Specialization
        typedef int8_t word;
    };
    template<>
    struct atomic_word<2> {          // Specialization
        typedef int16_t word;
    };
    template<>
    struct atomic_word<4> {          // Specialization
        typedef int32_t word;
    };
    /// #endif

    template<>
    struct atomic_word<8> {          // Specialization
        typedef int64_t word;
    };

    } } // namespace tbb::internal

typedef tbb::internal::atomic_word<__TBB_WORDSIZE>::word __TBB_word; // TODO: or is that intptr_t?

template<typename T>
inline T __TBB_load_with_acquire(const volatile T& location) {
    typedef typename tbb::internal::atomic_word<sizeof(T)>::word word;
    return T(tbb::internal::__TBB_Load<sizeof(T),word,tbb::acquire>::op(&location));
}

// TODO: instead of blindly casting away here (first implementation),
// do whatever is appropriate at the call site to make the second implementation sufficient?
#if 1
template<typename T, typename V>
inline void __TBB_store_with_release(volatile T& location, V value) {
    typedef typename tbb::internal::atomic_word<sizeof(T)>::word word;
    tbb::internal::__TBB_Op<sizeof(T),word,tbb::release,tbb::internal::__TBB_op_store>::op(&location, word(value));
}
#else
template<typename T>
inline void __TBB_store_with_release(volatile T& location, T value) {
    typedef typename tbb::internal::atomic_word<sizeof(T)>::word word;
    tbb::internal::__TBB_Op<sizeof(T),word,tbb::release,tbb::internal::__TBB_op_store>::op(&location, word(value));
}
#endif

#ifndef __TBB_Pause
    inline void __TBB_Pause(int32_t) {
        __TBB_Yield();
    }
#endif

namespace tbb { namespace internal {

// TODO: Why is this often used in a compare-and-swap loop
//     where the comparand is from the previous loop iteration?
//     It would seem that this makes the new iteration more
//     likely to fail because the underlying value has changed,
//     thus costing extra iterations that are ever more
//     likely to be interfered with. Maybe instead the
//     result from the previous iteration should be used
//     for only a few iterations and then that link should be
//     broken and a new value read just before the compare-and-swap,
//     but it is not clear when that transition should be?
//     Alternatively, delayed iterations should be alternated
//     with no-delay iterations (just one, or a few?),
//     and the next question would be whether those no-delay
//     iterations should be initiated from within AtomicBackoff
//     (probably not, because that makes little sense for lock
//     acquisition that is not based on compare-and-swap, if any),
//     or otherwise perhaps unrolled.
//! Class that implements exponential backoff.
/** See implementation of SpinwaitWhileEq for an example. */
class AtomicBackoff {
    //! Time delay, in units of "pause" instructions. 
    /** Should be equal to approximately the number of "pause" instructions
        that take the same time as an context switch. */
    static const int32_t LOOPS_BEFORE_YIELD = 16;
    int32_t count;
public:
    AtomicBackoff() : count(1) {}

    //! Pause for a while.
    void pause() {
        if( count<=LOOPS_BEFORE_YIELD ) {
            __TBB_Pause(count);
            // Pause twice as long the next time.
            count*=2;
        } else {
            // Pause is so long that we might as well yield CPU to scheduler.
            __TBB_Yield();
        }
    }

    // pause for a few times and then return false immediately.
    bool bounded_pause() {
         if( count<=LOOPS_BEFORE_YIELD ) {
             __TBB_Pause(count);
             // Pause twice as long the next time.
             count*=2;
             return true;
         } else {
             return false;
	 }
    }

    void reset() {
        count = 1;
    }
};

// __TBB_CompareAndSwap

    // utility (used with unsigned T to prevent any mishaps)
    template<size_t S, typename T, memory_semantics M>
    static inline T __TBB_MaskedCompareAndSwap (volatile void* const ptr, const T value, const T comparand ) {
        volatile uint32_t* const base = (volatile uint32_t*)( (uintptr_t)(ptr) & ~(uintptr_t)(0x3) ); // aligned to 4
        #if __TBB_BIG_ENDIAN
        const uint8_t bitoffset = ( (4-S) - ( (uint8_t *)ptr - (uint8_t *)base ) ) * 8;
        #elif defined(__TBB_BIG_ENDIAN) // defined but 0, meaning little-endian
        const uint8_t bitoffset =           ( (uint8_t *)ptr - (uint8_t *)base )   * 8;
        #else
        #error __TBB_BIG_ENDIAN unknown
        #endif
        const uint32_t mask = ( (1<<(S*8) ) - 1)<<bitoffset;
        uint32_t big_observed;
            if( tbb::relaxed==M || tbb::acquire==M ) {
                big_observed = uint32_t(__TBB_Load<4,int32_t,__TBB_NON_ACQUIRE_TO_RELAXED(M)>::op(base));
                const T observed = (T)( (big_observed & mask) >> bitoffset);
                if( comparand != observed ) return observed;
            }
            else {
                big_observed = *reinterpret_cast<volatile uint32_t*>(base);
                // TODO: what would it do to the semantics to give up if big_observed doesn't suit the comparand?
                // alternatively, the initial probe might be omitted (set at 0)
            }
        for( AtomicBackoff b; ; b.pause() ) {
            const uint32_t big_value     = ( big_observed & ~mask ) | ( ((uint32_t)value    ) << bitoffset );
            const uint32_t big_comparand = ( big_observed & ~mask ) | ( ((uint32_t)comparand) << bitoffset );
            big_observed = uint32_t(__TBB_CompareAndSwap<4,int32_t,M>::op( base, int32_t(big_value), int32_t(big_comparand) ));
            if( big_comparand==big_observed ) { // success
                return comparand; // avoids computing observed and returning that
            }
            const T observed = (T)( (big_observed & mask) >> bitoffset);
            if( comparand!=observed) { // failure
                return observed;
            }
        }
    }

    #define __TBB_M(B)                                                                                 \
        template<size_t S, memory_semantics M> struct __TBB_CompareAndSwap<S,int##B##_t,M> {           \
            static inline int##B##_t op (volatile void *ptr, int##B##_t value, int##B##_t comparand) { \
                return int##B##_t(__TBB_MaskedCompareAndSwap<S,uint##B##_t,M>(                         \
                    ptr, uint##B##_t(value), uint##B##_t(comparand)                                    \
                    ));                                                                                \
            }                                                                                          \
        };
    __TBB_M(8)
    __TBB_M(16)
    #undef __TBB_M

// __TBB_FetchOp

    template<size_t S, typename T, memory_semantics M, typename atomic_op>
    inline T __TBB_FetchOp<S,T,M,atomic_op>::op( volatile void *ptr, T arg ) {
        T observed = *reinterpret_cast<volatile T *>(ptr);
        for( AtomicBackoff b; ; b.pause() ) for( int i = 0; i<2; ++i ) {
            const T previous = observed;
            if( previous==(observed = __TBB_CompareAndSwap<S,T,M>::op( ptr, atomic_op::op(previous, arg), previous )) )
                return observed;
        }
    }

// Macro __TBB_TypeWithAlignmentAtLeastAsStrict(T) should be a type with alignment at least as 
// strict as type T.  Type type should have a trivial default constructor and destructor, so that
// arrays of that type can be declared without initializers.  
// It is correct (but perhaps a waste of space) if __TBB_TypeWithAlignmentAtLeastAsStrict(T) expands
// to a type bigger than T.
// The default definition here works on machines where integers are naturally aligned and the
// strictest alignment is 16.
#ifndef __TBB_TypeWithAlignmentAtLeastAsStrict

#if __GNUC__ || __SUNPRO_CC
struct __TBB_machine_type_with_strictest_alignment {
    int member[4];
} __attribute__((aligned(16)));
#elif _MSC_VER
__declspec(align(16)) struct __TBB_machine_type_with_strictest_alignment {
    int member[4];
};
#else
#error Must define __TBB_TypeWithAlignmentAtLeastAsStrict(T) or __TBB_machine_type_with_strictest_alignment
#endif

template<size_t N> struct type_with_alignment {__TBB_machine_type_with_strictest_alignment member;};
template<> struct type_with_alignment<1> { char member; };
template<> struct type_with_alignment<2> { uint16_t member; };
template<> struct type_with_alignment<4> { uint32_t member; };
template<> struct type_with_alignment<8> { uint64_t member; };

#if _MSC_VER||defined(__GNUC__)&&__GNUC__==3 && __GNUC_MINOR__<=2  
//! Work around for bug in GNU 3.2 and MSVC compilers.
/** Bug is that compiler sometimes returns 0 for __alignof(T) when T has not yet been instantiated.
    The work-around forces instantiation by forcing computation of sizeof(T) before __alignof(T). */
template<size_t Size, typename T> 
struct work_around_alignment_bug {
#if _MSC_VER
    static const size_t alignment = __alignof(T);
#else
    static const size_t alignment = __alignof__(T);
#endif
};
#define __TBB_TypeWithAlignmentAtLeastAsStrict(T) tbb::internal::type_with_alignment<tbb::internal::work_around_alignment_bug<sizeof(T),T>::alignment>
#elif __GNUC__ || __SUNPRO_CC
#define __TBB_TypeWithAlignmentAtLeastAsStrict(T) tbb::internal::type_with_alignment<__alignof__(T)>
#else
#define __TBB_TypeWithAlignmentAtLeastAsStrict(T) __TBB_machine_type_with_strictest_alignment
#endif
#endif  /* ____TBB_TypeWithAlignmentAtLeastAsStrict */

} } // namespace tbb::internal

static inline __TBB_word __TBB_FetchAndAddWrelease(volatile void* ptr, __TBB_word arg) {
    return tbb::internal::__TBB_FetchOp<__TBB_WORDSIZE,__TBB_word,tbb::release,tbb::internal::__TBB_op_add>::op(ptr, arg);
}
static inline __TBB_word __TBB_FetchAndIncrementWacquire(volatile void* ptr) {
    return tbb::internal::__TBB_FetchOp<__TBB_WORDSIZE,__TBB_word,tbb::acquire,tbb::internal::__TBB_op_add>::op(ptr, 1);
}
static inline __TBB_word __TBB_FetchAndDecrementWrelease(volatile void* ptr) {
    return tbb::internal::__TBB_FetchOp<__TBB_WORDSIZE,__TBB_word,tbb::release,tbb::internal::__TBB_op_add>::op(ptr, -1);
}

#ifndef __TBB_Log2
inline intptr_t __TBB_Log2( uintptr_t x ) {
    long result = -1;
    for(; x; x>>=1 ) ++result;
    return result;
}
#endif

// TODO: why no AtomicBackoff here before? does it matter that there is one now?
// TODO: build on top of tbb::atomic instead?
// TODO: what should be the memory semantics (ordered chosen for safety, but...)
inline void __TBB_AtomicAND( volatile void *operand, uintptr_t arg ) {
    typedef tbb::internal::atomic_word<sizeof(uintptr_t)>::word word;
    tbb::internal::__TBB_Op<sizeof(uintptr_t),word,tbb::ordered,tbb::internal::__TBB_op_and>::op( operand, word(arg) );
}
inline void __TBB_AtomicOR ( volatile void *operand, uintptr_t arg ) {
    typedef tbb::internal::atomic_word<sizeof(uintptr_t)>::word word;
    tbb::internal::__TBB_Op<sizeof(uintptr_t),word,tbb::ordered,tbb::internal::__TBB_op_or >::op( operand, word(arg) );
}

// TODO: what memory semantics?
#ifndef __TBB_TryLockByte
inline bool __TBB_TryLockByte( unsigned char &flag ) {
    return tbb::internal::__TBB_CompareAndSwap<1,int8_t,tbb::ordered>::op(&flag,1,0)==0;
}
#endif

#ifndef __TBB_LockByte
inline uintptr_t __TBB_LockByte( unsigned char& flag ) {
    if ( !__TBB_TryLockByte(flag) ) {
        tbb::internal::AtomicBackoff b;
        do {
            b.pause();
        } while ( !__TBB_TryLockByte(flag) );
    }
    return 0;
}
#endif

#endif /* __TBB_machine_H */
